{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: ObjStack.Pas
|
| Description: Stack container.
|
| History: Version 1.0  March 1996. Copyright (c) 1996 Michel Brazeau
|                                   Interval Software
|
|---------------------------------------------------------------------------}
unit ObjStack;

{$I Contain.Inc}

interface

uses
    ObjPerst, { TObjectReader, TObjectWriter }
    ObjBase,  { TContainer }
    ObjList;  { TUnOrderedList }

type

TStack = class(TContainer)
protected
    List : TUnOrderedList;

    function GetSize : LongInt; override;

    function QueryEmpty : Boolean; override;

    procedure SetOwnership( const InOwnership : Boolean ); override;

public
    constructor Create(const InObjectType : TClass);

    { destroys calls Clear to remove and free every element in the stack }
    destructor Destroy; override;

    { clear all the items in the stack. Free is called for each item }
    procedure Clear; override;

    { push an element on the stack }
    procedure Push(const Obj : TObject);

    { pop the most recent element pushed, May raise EContainerIsEmpty }
    function Pop : TObject;

    { peek at the top element on the stack, May raise EContainerIsEmpty }
    function PeekTop : TObject;

    { for each item in the container, call ProcIterator. ProcIterator is
      a file level procedure. }
    procedure ForEachCallProc( const ProcIterator : TProcIterator;
                               const CallBackData : LongInt); override;

    { for each item in the container, call MethodIterator. MethodIterator is
      a method of a class. }
    procedure ForEachCallMethod( const MethodIterator : TMethodIterator;
                                 const CallBackData   : LongInt ); override;

end;

{ stream registration routines. }
procedure ReadStack( const Obj    : TObject;
                     const Reader : TObjectReader ); far;

procedure WriteStack( const Obj    : TObject;
                      const Writer : TObjectWriter ); far;

{--------------------------------------------------------------------------}

implementation

uses
    ObjType,  { RaiseXXX }
    ObjCmp;   { CompareEqual }

{--------------------------------------------------------------------------}

constructor TStack.Create(const InObjectType : TClass);
begin
    inherited Create(InObjectType);

    List := TUnOrderedList.Create(InObjectType, CompareEqual);

    { MB May 13, 1997. Set the base class FCompareFunc for streaming purposes }
    FCompareFunc := CompareEqual;
end;

{--------------------------------------------------------------------------}

destructor TStack.Destroy;
begin
    { called inherited Destroy first, so when Clear is called, List is
      still valid }
    inherited Destroy;

    List.Free;
end;

{--------------------------------------------------------------------------}

procedure TStack.SetOwnership( const InOwnership : Boolean );
begin
    { maintain the FOwnership data member in the base class }
    inherited SetOwnership(InOwnership);

    List.Ownership := InOwnership;
end;

{--------------------------------------------------------------------------}

procedure TStack.Clear;
begin
    { check for nil, as Clear may be invoked on a partially constructed
      object }
    if List <> nil then
        List.Clear;
end;

{--------------------------------------------------------------------------}

procedure TStack.Push(const Obj : TObject);
begin
    List.InsertAtHead(Obj);
end;

{--------------------------------------------------------------------------}

function TStack.Pop : TObject;
begin
    if List.IsEmpty then
        RaiseContainerIsEmpty;

    List.GotoFirst;

    { delete the node, but don't free the object in the node, as this is
      the object being returned ! }
    Result := List.DetachCurrent;
end;

{--------------------------------------------------------------------------}

function TStack.PeekTop : TObject;
begin
    if List.IsEmpty then
        RaiseContainerIsEmpty;

    List.GotoFirst;

    Result := List.CurrentObj;
end;

{--------------------------------------------------------------------------}

function TStack.GetSize : LongInt;
begin
    Result := List.Size;
end;

{--------------------------------------------------------------------------}

function TStack.QueryEmpty : Boolean;
begin
    Result := List.IsEmpty;
end;

{--------------------------------------------------------------------------}

procedure TStack.ForEachCallProc( const ProcIterator : TProcIterator;
                                  const CallBackData : LongInt );
begin
    List.ForEachCallProc(ProcIterator, CallBackData);
end;

{--------------------------------------------------------------------------}

procedure TStack.ForEachCallMethod( const MethodIterator : TMethodIterator;
                                    const CallBackData   : LongInt);
begin
    List.ForEachCallMethod(MethodIterator, CallBackData);
end;

{--------------------------------------------------------------------------}

procedure ReadStack( const Obj    : TObject;
                     const Reader : TObjectReader );
var
    Stack : TStack;
begin
    Stack := Obj as TStack;

    ReadContainer(Obj, Reader);

    Stack.List := Reader.ReadChildObject as TUnorderedList;
end;

{--------------------------------------------------------------------------}

procedure WriteStack( const Obj    : TObject;
                      const Writer : TObjectWriter );
var
    Stack : TStack;
begin
    Stack := Obj as TStack;

    WriteContainer(Stack, Writer);

    Writer.WriteChildObject( Stack.List );
end;

{---------------------------------------------------------------------}

initialization

    {$ifdef Win32}
    RegisterStreamable( TStack,
                        [TObjectReadProc(@ReadStack)],
                        [TObjectWriteProc(@WriteStack)] );
    {$else}
    RegisterStreamable( TStack,
                        [ReadStack],
                        [WriteStack] );
    {$endif}

end.
