{----------------------------------------------------------------------
|
| Library: Spider Object Database
|
| Module: OdbBde.Pas
|
| Description: BDE object database driver class.
|
| Notes:
|        With Paradox, transactions (rollback) does not work with SQL
|        statements
|
| History: Jan 03, 1998. Michel Brazeau, first version
|          May 07, 1998. When DBase tables, indices are properly defined
|                        using expression indices.
|          Sep 14, 1999. Add support in driver for TObjectQuery datasets.
|          Mar 10, 2001. MB override introspection methods and testcase
|
|---------------------------------------------------------------------}
unit OdbBde;

{$I Odb.Inc}

interface

uses
    {$ifdef __Testing}
    DriverTest, { for TOdbDriverTest }
    {$endif}

    Classes,  { for TMemoryStream }
    SysUtils, { for Exception, StrPas }
    Db,       { for TDatabase, TDataset }
    DbTables, { for TTable }
    ObjList,  { for TUnOrderedList }
    Odb,      { for TObjectFile, TOdbDriver }
    OdbIndex, { for TOdbIndex }
    OdbField; { for TOdbField }

type

TOdbBdeTable = class(TTable)
public
    { for Paradox table, the primary index name must be '' when talking to
      BDE. PrimaryIndexName is the index name as defined by the related
      object file. }
    PrimaryIndexName : String;
end;

{--------------------------------------------------------------------}

TBdeOdbDriver = class(TOdbDriver)
protected

    FDatabase      : TDatabase;

    FTableType     : TTableType;

    FSessionName   : String;

    FSqlBased      : Boolean;

public

    { returns True if the underlying BDE database connection is an
      SQL based database. This property is only valid when the database
      has been opened. }
    property SqlBased : Boolean read FSQLBased;

    constructor Create(const InDatabaseName : String); override;
    destructor Destroy; override;

    procedure InternalStartTransaction; override;
    procedure InternalCommit; override;
    procedure InternalRollback; override;

    procedure InternalOpenDatabase(
                            const ServerName : String;
                            const UserName   : String;
                            const Password   : String
                          ); override;

    procedure InternalCloseDatabase; override;

    function InternalOpenDataset( const ObjectDataset : TObjectDataset ) : TOdbHandle; override;

    function InternalOpenExclusive( const ObjectFile : TObjectFile ) : TOdbHandle; override;

    procedure InternalCloseDataset( const Handle : TOdbHandle); override;

    procedure InternalDeleteDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalCreateDataset( const ObjectDataset : TObjectDataset ); override;

    procedure InternalInsert(
                      const Handle  : TOdbHandle;
                      const Fields  : TUnOrderedList;
                      const Stream  : TMemoryStream ); override;

    procedure InternalUpdate(
                      const Handle     : TOdbHandle;
                      const Fields     : TUnOrderedList;
                      const Stream     : TMemoryStream ); override;

    procedure InternalDelete( const Handle : TOdbHandle ); override;

    function InternalCreateObjectUpdate : TOdbHandle; override;
    procedure InternalDestroyObjectUpdate( const Handle : TOdbHandle ); override;

    procedure InternalInsertSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); override;

    procedure InternalUpdateSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String ); override;

    procedure InternalDeleteSQL( const Handle      : TOdbHandle;
                                 const ORMapping   : TORMapping;
                                 const DatasetName : String ); override;

    function InternalGetIndexName( const Handle : TOdbHandle ) : String; override;

    procedure InternalSetIndexName(
                            const Handle    : TOdbHandle;
                            const IndexName : String ); override;

    procedure InternalGetCurrent(
                          const Handle  : TOdbHandle;
                          const Stream  : TMemoryStream ); override;

    function InternalLockCurrent( const Handle : TOdbHandle ) : Boolean; override;

    procedure InternalUnLockCurrent( const Handle : TOdbHandle ); override;

    function InternalFindNext( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindPrior( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindFirst( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindLast( const Handle : TOdbHandle ) : Boolean; override;

    function InternalFindExact(
                        const Handle : TOdbHandle;
                        const Fields : TUnOrderedList;
                        const Index  : TOdbIndex ) : Boolean; override;

    procedure InternalFindNearest(
                          const Handle : TOdbHandle;
                          const Fields : TUnOrderedList;
                          const Index  : TOdbIndex ); override;

    function InternalGetBookmark( const Handle : TOdbHandle ) : Pointer; override;
    procedure InternalGotoBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;
    procedure InternalFreeBookmark( const Handle   : TOdbHandle;
                                    const Bookmark : Pointer ); override;

    function InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt; override;

    function InternalIsEmpty( const Handle : TOdbHandle ) : Boolean; override;

    procedure InternalPack( const Handle : TOdbHandle ); override;

    function GetSupportedFieldTypes : TOdbFieldTypes; override;

    function CheckDuplicateInsert : Boolean; override;
    function CheckDuplicateUpdate : Boolean; override;

    function GetBdeDataset( const Handle : TOdbHandle ) : Db.TDataset;

    function GetDataAccessMethods : TDataAccessMethodSet; override;


    function AutoIncSupported              : Boolean; override;
    function IntegerIndexSupported         : Boolean; override;
    function FloatIndexSupported           : Boolean; override;
    function CaseInsensitiveIndexSupported : Boolean; override;
    function DescendingIndexSupported      : Boolean; override;
    function TransactionsSupported         : Boolean; override;
    function FileLockingSupported          : Boolean; override;
    function ObjectLockingSupported        : Boolean; override;
    function SingleIndexOnly               : Boolean; override;

protected
    function BdeOpenFile( const ObjectFile : TObjectFile;
                          const Exclusive  : Boolean ) : TOdbHandle;

    function BdeOpenQuery( const ObjectQuery : TObjectQuery ) : TOdbHandle;

    procedure OpenTableForObjectUpdate( const Handle      : TOdbHandle;
                                        const ORMapping   : TORMapping;
                                        const DatasetName : String );

    procedure CreateTable( const ObjectDataset : TObjectDataset );

    procedure CreateTableWithSQL( const ObjectDataset : TObjectDataset );

    procedure DeleteTable( const ObjectDataset : TObjectDataset );

    procedure DeleteTableWithSQL( const ObjectDataset : TObjectDataset );

end;

EBdeTableNotOpenedExclusively    = class(Exception);


{$ifdef __Testing}

type

TParadoxBDEOdbDriverTest = class( TOdbDriverTest )
protected
     procedure GetParams( var DriverClass                   : TOdbDriverClass;
                          var DatabaseName                  : String;
                          var ServerName                    : String;
                          var UserName                      : String;
                          var Password                      : String
                        ); override;
end;

{ Interbase BDE }
TInterbaseBDEOdbDriverTest = class( TOdbDriverTest )
protected
     procedure GetParams( var DriverClass                   : TOdbDriverClass;
                          var DatabaseName                  : String;
                          var ServerName                    : String;
                          var UserName                      : String;
                          var Password                      : String
                        ); override;
end;

{$endif}


{--------------------------------------------------------------------}

implementation

uses
    ObjPerst,
    FileCtrl, { for DirectoryExists }
    {$ifdef Ver80}
    DbiTypes, { dbiGetRecord }
    DbiProcs, { dbiWriteLock }
    DbiErrs,  { ERRCODE_XXX }
    {$else}
    Bde,      { dbiGetRecord }
    {$endif}
    ObjBuckt, { for TStringCombo }
    OdbMisc;  { for RaiseOdbException }

const

CObjectFieldName = 'Object';

{--------------------------------------------------------------------}

procedure DbiCheck(Status : DbiResult);
begin
    {$ifdef __Delphi3AndAbove}
    DbTables.Check(Status);
    {$else}
    Db.Check(Status);
    {$endif}
end;

{--------------------------------------------------------------------}

constructor TBdeOdbDriver.Create(const InDatabaseName : String);
begin
    FTableType            := ttDefault;

    FDatabase             := TDatabase.Create(nil);

    if not DirectoryExists(InDatabaseName) then
        FDatabase.AliasName      := InDatabaseName;

    FDatabase.DatabaseName   := InDatabaseName;

    { MB Mar 10, 2001. Only use dirty read on local BDE databases }
    if not FDatabase.IsSQLBased then
        FDatabase.TransIsolation := tiDirtyRead;

    inherited Create(InDatabaseName);
end;

{--------------------------------------------------------------------}

destructor TBdeOdbDriver.Destroy;
begin
    { invoke inherited first, to close the database if it is opened before
      invoking FDatabase.Free }
    inherited Destroy;

    FDatabase.Free;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalStartTransaction;
begin
    FDatabase.StartTransaction;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalCommit;
begin
    FDatabase.Commit;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalRollback;
begin
    FDatabase.Rollback;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalOpenDatabase(
                        const ServerName : String;
                        const UserName   : String;
                        const Password   : String
                      );
var
    PropValue : array[0..80] of Char;
    Length    : Word;
begin
    FDatabase.KeepConnection := True;
    FDatabase.LoginPrompt    := False;
    FDatabase.Params.Values['SERVER NAME'] := ServerName;
    FDatabase.Params.Values['USER NAME']   := UserName;
    FDatabase.Params.Values['PASSWORD']    := Password;

    { MB Mar 11, 2001. Use a session component, too allow creating two
      TObjectDatabase's on the same database }
    FSessionName := 'ODB' + IntToStr(Sessions.Count + 1);
    Sessions.OpenSession(FSessionName);
    FDatabase.SessionName  := FSessionName;

    FDatabase.Open;

    { By default, use Paradox tables if the database name is a directory. }
    if DirectoryExists(DatabaseName) then
       FTableType := ttParadox;

    FillChar(PropValue, SizeOf(PropValue), 0);
    DbiGetProp( hDBIObj(FDatabase.Handle), dbDATABASETYPE, @PropValue,
                SizeOf(PropValue), Length );
    if StrPas(PChar(@PropValue)) = 'STANDARD' then
    begin
        FillChar(PropValue, SizeOf(PropValue), 0);
        DbiGetProp( hDBIObj(FDatabase.Handle), dbDEFAULTDRIVER , @PropValue,
                    SizeOf(PropValue), Length );
        if StrPas(PChar(@PropValue)) = 'DBASE' then
            FTableType := ttDBase
        else if StrPas(PChar(@PropValue)) = 'PARADOX' then
            FTableType := ttParadox;
    end;

    FSqlBased := FDatabase.IsSqlBased;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalCloseDatabase;
var
    Session : TSession;
begin
    FDatabase.Close;

    Session := Sessions.FindSession(FSessionName);
    if Session <> nil then
        Session.Free;
    FSessionName := '';
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalCreateObjectUpdate : TOdbHandle;
var
    Dataset : TDbDataset;

begin
    Dataset := nil;
    try
        { MB Mar 21, 2001. With Local database use TTable objects to update the
          database, as sql insert and update are not supported with autoincrement
          fields. Even with  }
        if SqlBased then
            Dataset := TQuery.Create(nil)
        else
        begin
            Dataset := TOdbBdeTable.Create(nil);

            TOdbBdeTable(Dataset).TableType        := FTableType;
        end;

        Dataset.DatabaseName := FDatabase.DatabaseName;
        Dataset.SessionName  := FDatabase.SessionName;

        Result := NewHandle(Dataset);
    except
        Dataset.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalDestroyObjectUpdate( const Handle : TOdbHandle );
begin
    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.BdeOpenFile( const ObjectFile : TObjectFile;
                                    const Exclusive  : Boolean ) : TOdbHandle;
var
    Table : TOdbBdeTable;
    Index : TOdbIndex;
    Field : TMemoField;

begin
    Table := TOdbBdeTable.Create(nil);
    try
        Table.DatabaseName := FDatabase.DatabaseName;
        Table.SessionName  := FDatabase.SessionName;
        Table.TableName    := ObjectFile.DatasetName;
        Table.TableType    := FTableType;
        Table.Exclusive    := Exclusive;
        Table.Open;

        Result := NewHandle(Table);
    except
        Table.Free;
        raise;
    end;

    if Table.TableType = ttParadox then
    begin
        if ObjectFile.Indices.GotoFirst then
        repeat
            Index := (ObjectFile as TObjectFile).Indices.CurrentObj as TOdbIndex;

            if iaPrimary in Index.Attributes then
            begin
                Table.PrimaryIndexName := Index.Name;
                break;
            end;
        until not ObjectFile.Indices.GotoNext;
    end;

    { See the notes in InternalCreateTable is to why the Object field
      is a TMemoField rather than a TBlobField. }
    if Table.TableType = ttDBase then
    begin
        Field := (Table.FieldByName(CObjectFieldName) as TMemoField);
        Field.Transliterate := False;
    end;
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.BdeOpenQuery(
                           const ObjectQuery : TObjectQuery ) : TOdbHandle;
var
    Query     : TQuery;

begin
    Query     := nil;
    try
        Query     := TQuery.Create(nil);

        Query.DatabaseName := FDatabase.DatabaseName;
        Query.SessionName  := FDatabase.SessionName;
        Query.SQL.Assign( ObjectQuery.Select );
        Query.Open;

        Result := NewHandle(Query);
    except
        Query.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalOpenDataset(
                       const ObjectDataset : TObjectDataset ) : TOdbHandle;
begin
    if ObjectDataset is TObjectFile then
        Result := BdeOpenFile(TObjectFile(ObjectDataset), False)
    else if ObjectDataset is TObjectQuery then
    begin
        if TObjectQuery(ObjectDataset).Select.Count = 0 then
            RaiseOdbException( EOdbSelectStatementNotSpecified,
                               ObjectDataset.DatasetName );

        Result := BdeOpenQuery( TObjectQuery(ObjectDataset) );
    end
    else
    begin
        Result := 0; { prevent compiler warning }
        RaiseOdbException(EOdbError, 'Internal error ' + ObjectDataset.DatasetName);
    end;
end;
{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalOpenExclusive(
                                  const ObjectFile : TObjectFile ) : TOdbHandle;
begin
    Result := BdeOpenFile(ObjectFile, True);
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalCloseDataset( const Handle : TOdbHandle);
var
    Dataset : Db.TDataset;
begin
    Dataset := GetBdeDataset(Handle);

    Dataset.Close;

    ReleaseHandle(Handle);
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.DeleteTable(
                          const ObjectDataset : TObjectDataset );
var
    Table : TOdbBdeTable;
begin
    Table := TOdbBdeTable.Create(nil);
    try
        Table.DatabaseName := FDatabase.DatabaseName;
        Table.SessionName  := FDatabase.SessionName;
        Table.TableName    := ObjectDataset.DatasetName;
        Table.TableType    := FTableType;
        Table.DeleteTable;
    finally
        Table.Free;
    end;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.DeleteTableWithSQL(
                          const ObjectDataset : TObjectDataset );
var
    Query     : TQuery;
begin
    Query     := TQuery.Create(nil);
    try
        Query.DatabaseName := FDatabase.DatabaseName;
        Query.SessionName  := FDatabase.SessionName;

        Query.SQL.Add('drop table ' + ObjectDataset.DatasetName );
        Query.ExecSQL;

    finally
        Query.Free;
    end;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalDeleteDataset( const ObjectDataset : TObjectDataset );
begin
    if ObjectDataset is TObjectFile then
        DeleteTable( ObjectDataset)
    else
        DeleteTableWithSQL( ObjectDataset );
end;

{--------------------------------------------------------------------}

{ add Fields to Table }
procedure AddFields( const Fields  : TUnOrderedList;
                     const Table   : TOdbBdeTable );
var
    Field   : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftString  : Table.FieldDefs.Add(
                                 Field.Name,
                                 ftString, Field.Size-1, False );
            oftInteger : Table.FieldDefs.Add(
                                 Field.Name,
                                 ftInteger, 0, False );
            oftFloat   : Table.FieldDefs.Add(
                                 Field.Name,
                                 ftFloat, 0, False );
            {$ifndef Ver80}
            oftAutoInc : Table.FieldDefs.Add(
                                 Field.Name,
                                 ftAutoInc, 0, False );
            {$endif}

        end;
    until not Fields.GotoNext;

    { Blob fields must be created as memo fields to circumvent a bug
      occuring with DBase tables. TBlobFields have the Transliterate
      property set True by default, and not public, so corruption was
      occuring because "raw" blob fields were being translated. Since
      the TMemoField class has published the Transliterate property, a
      MemoField is used, and the Transliterate property will be set
      False when the table is opened. The TMemoField is an immediate
      descendant of a TBlobField, so this does not create any problems. }
    if Table.TableType = ttDBase then
        Table.FieldDefs.Add(CObjectFieldName, ftMemo, 0, False)
    else
        Table.FieldDefs.Add(CObjectFieldName, ftBlob, 0, False);
end;

{--------------------------------------------------------------------}

{ convert Spider Odb index attributes to VCL index options }
function IndexAttributesToOptions(
            const Attributes : TOdbIndexAttributes ) : TIndexOptions;
begin
    Result := [];

    if iaUnique in Attributes then
        Result := Result + [ixUnique];

    if iaCaseInsensitive in Attributes then
        Result := Result + [ixCaseInsensitive];

    if iaDescending in Attributes then
        Result := Result + [ixDescending];

    if iaPrimary in Attributes then
        Result := Result + [ixPrimary];
end;

{--------------------------------------------------------------------}

procedure AddIndices( const Indices          : TUnOrderedList;
                      const FieldList        : TUnOrderedList;
                      const Table            : TOdbBdeTable;
                      const Physical         : Boolean );
var
    Index        : TOdbIndex;
    IndexOptions : TIndexOptions;
    FieldNames   : String;
    IndexName    : String;

begin
    if Indices.GotoFirst then
    repeat
        Index        := Indices.CurrentObj as TOdbIndex;
        FieldNames   := Index.FieldNamesAsString(';');
        IndexOptions := IndexAttributesToOptions(Index.Attributes);

        if (Table.TableType = ttDBase) then
        begin
            FieldNames := IndexToXBaseIndexExp(
                               Index, FieldList,
                               11, 20, 4 );
            IndexOptions := IndexOptions + [ixExpression];
        end;

        IndexName := Index.Name;

        if (Table.TableType = ttParadox) and
           (ixPrimary in IndexOptions) then
            IndexName := ''; { Paradox must have primary index = '' }

        if (Table.TableType <> ttParadox) then
        begin
            { primary indexes only apply to Paradox. A primary index is
              similar to a unique index, so replace primary by unique. }
            if ixPrimary in IndexOptions then
                IndexOptions := IndexOptions - [ixPrimary] + [ixUnique];
        end;

        if Table.TableType = ttDBase then
        begin
            { case insensitive indices do not apply to DBase. }
            IndexOptions := IndexOptions - [ixCaseInsensitive];

            { A unique index in DBase is very confusing. It does not mean
              unique values, but rather duplicates are ignored when
              navigating. This was creating a serious problem with soft
              delete DBase records. Once a record was deleted it is not
              possible to reinsert a record with the same key, as the BDE
              engine seems to treat the new record as a duplicate of the
              first record, and the new record is not navigable. The workaround
              is not to set the ixUnique option, but to check for duplicates
              manually. See CheckDuplicateInsert, CheckDuplicateUpdate.
              Jan 25, 1998. MB I've just downloaded BDE 4.51, and the version
              is said to support "unique" indices in the proper sense of the
              word, but for backward compatibility the said approached is
              used. }
            IndexOptions := IndexOptions - [ixUnique];
        end;

        if Physical then
            Table.AddIndex( IndexName,
                            FieldNames,
                            IndexOptions )
        else
            Table.IndexDefs.Add( IndexName,
                                 FieldNames,
                                 IndexOptions );

    until not Indices.GotoNext;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalCreateDataset(
                          const ObjectDataset : TObjectDataset
                        );
begin
    if ObjectDataset is TObjectFile then
        CreateTable(ObjectDataset)
    else
    begin
        if FDatabase.IsSQLBased then
            CreateTableWithSQL(ObjectDataset)
        else
            { with Local SQL (Paradox and DBase), the Create Index
              SQL statements fail }
            CreateTable(ObjectDataset);
    end;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.CreateTable(
                      const ObjectDataset : TObjectDataset
                    );
var
    Table : TOdbBdeTable;

begin
    Table := TOdbBdeTable.Create(nil);
    try
        Table.DatabaseName := FDatabase.DatabaseName;
        Table.TableName    := ObjectDataset.DatasetName;
        Table.TableType    := FTableType;
        Table.SessionName  := FSessionName;

        Table.FieldDefs.Clear;
        Table.IndexDefs.Clear;

        AddFields(ObjectDataset.Fields, Table);

        if (Table.TableType = ttDBase) then
        begin
            Table.CreateTable;
            { indices must be added individual to properly handle DBase expression
              indices. }
            AddIndices(ObjectDataset.Indices, ObjectDataset.Fields, Table, True);
        end
        else
        begin
            AddIndices(ObjectDataset.Indices, ObjectDataset.Fields, Table, False);
            Table.CreateTable;
        end;

    finally
        Table.Free;
    end;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.CreateTableWithSQL(
                          const ObjectDataset : TObjectDataset );
var
    Statement : TStringList;
    Query     : TQuery;
    LineNo    : Integer;
    BlobType  : String;
begin
    Statement := TStringList.Create;
    Query     := TQuery.Create(nil);
    try
        Query.DatabaseName := FDatabase.DatabaseName;
        Query.SessionName  := FSessionName;

        if not FDatabase.IsSQLBased then
            BlobType := 'blob(128,2)'
        else
            { this is the propery default for Interbase }
            BlobType := 'blob';

        BuildCreateTableStatement( Statement, ObjectDataset.ORMapping,
                                   ObjectDataset.DatasetName,
                                   'character', 'integer', 'float', BlobType);
        Query.SQL.Assign(Statement);
        Query.ExecSQL;

        BuildCreateIndexStatements( Statement, ObjectDataset );
        for LineNo := 0 to (Statement.Count-1) do
        begin
            Query.SQL.Clear;
            Query.SQL.Add( Statement[LineNo] );
            Query.ExecSQL;
        end;

    finally
        Query.Free;
        Statement.Free;
    end;
end;

{--------------------------------------------------------------------}

{ set the field values in Table given a list of TOdbField's }
procedure SetDatasetFieldValues( const Fields  : TUnOrderedList;
                                 const Dataset   : Db.TDataset );
var
    Field : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        { MB. Sep 14, 1998. Don't set autoinc field value. With the BDE shipped
          with Delphi 4, this raised an exception. }
        case Field.FieldType of
            oftString  : Dataset.FieldByName(Field.Name).AsString  := Field.AsString;
            oftInteger : Dataset.FieldByName(Field.Name).AsInteger := Field.AsInteger;
            oftFloat   : Dataset.FieldByName(Field.Name).AsFloat   := Field.AsFloat;
        end;
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------}

{ set the field values in Table for the fields belonging to Index }
procedure SetTableFieldValuesOnIndex( const Fields  : TUnOrderedList;
                                      const Index   : TOdbIndex;
                                      const Table   : TOdbBdeTable );
var
    Field       : TOdbField;
    StringCombo : TStringCombo;
begin
    if Index.FieldNames.GotoFirst then
    repeat
        StringCombo := Index.FieldNames.CurrentObj as TStringCombo;

        Field := GetFieldInList( StringCombo.Str, Fields );

        case Field.FieldType of
            oftString  : Table.FieldByName(Field.Name).AsString  := Field.AsString;
            oftInteger,
            oftAutoInc : Table.FieldByName(Field.Name).AsInteger := Field.AsInteger;
            oftFloat   : Table.FieldByName(Field.Name).AsFloat   := Field.AsFloat;
        end;
    until not Index.FieldNames.GotoNext;
end;

{--------------------------------------------------------------------}

{ Set the value of the autoincrement fields in Fields from the values
  contained in Table. }
procedure SetOdbAutoIncFieldValues( const Fields  : TUnOrderedList;
                                    const Table   : TOdbBdeTable );
var
    Field : TOdbField;
begin
    if Fields.GotoFirst then
    repeat
        Field := Fields.CurrentObj as TOdbField;

        case Field.FieldType of
            oftAutoInc : Field.AsInteger := Table.FieldByName(Field.Name).AsInteger;
        end;
    until not Fields.GotoNext;
end;

{--------------------------------------------------------------------}

procedure InsertInDataset( const Dataset : Db.TDataset;
                           const Fields  : TUnOrderedList;
                           const Stream  : TMemoryStream );
var
    Field   : TBlobField;
begin
    Dataset.Insert;
    try
        SetDatasetFieldValues(Fields, Dataset);

        Field := (Dataset.FieldByName(CObjectFieldName) as TBlobField);

        Field.LoadFromStream(Stream);

        Dataset.Post;

        { Update any autoincrement field in Fields. }
        if Dataset is TOdbBdeTable then
             SetOdbAutoIncFieldValues(Fields, TOdbBdeTable(Dataset));
    except
        Dataset.Cancel;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalInsert(
                  const Handle  : TOdbHandle;
                  const Fields  : TUnOrderedList;
                  const Stream  : TMemoryStream );

var
    Dataset : Db.TDataset;
begin
    Dataset := GetBdeDataset(Handle);

    InsertInDataset(Dataset, Fields, Stream);
end;

{--------------------------------------------------------------------}

procedure UpdateCurrentInDataset( const Dataset : Db.TDataset;
                                  const Fields  : TUnOrderedList;
                                  const Stream  : TMemoryStream );
begin
     try
        SetDatasetFieldValues(Fields, Dataset);
        (Dataset.FieldByName(CObjectFieldName) as TBlobField).LoadFromStream(Stream);
        Dataset.Post;
    except
        Dataset.Cancel;
        raise;
    end;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalUpdate(
                  const Handle     : TOdbHandle;
                  const Fields     : TUnOrderedList;
                  const Stream     : TMemoryStream );

var
    Dataset       : Db.TDataset;
begin
    Dataset       := GetBdeDataset(Handle);

   { MB May 21, 1998. TOdbBdeTable.Edit will attempt to lock the record before
      updating it. If the record has already been locked by the current session,
      it is acceptable to perform an update. A try,except block is used to
      attempt to unlock the object. If the object is locked by another session,
      the second call to Table.Edit will raise the exception. }
    try
        Dataset.Edit;
    except
        on EDBEngineError do
        begin
            if Dataset is TOdbBdeTable then
            begin
                { attempt the unlock only when using TTable objects }
                InternalUnLockCurrent(Handle);
                Dataset.Edit;
            end
            else
                raise;
        end;
    end;

    UpdateCurrentInDataset(Dataset, Fields, Stream);

end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalDelete( const Handle : TOdbHandle );
var
    Dataset       : Db.TDataset;
begin
    Dataset       := GetBdeDataset(Handle);

    Dataset.Delete
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.OpenTableForObjectUpdate(
                  const Handle      : TOdbHandle;
                  const ORMapping   : TORMapping;
                  const DatasetName : String );

var
    Table       : TOdbBdeTable;
begin
    Table := GetBdeDataset(Handle) as TOdbBdeTable;
    if not Table.Active then
    begin
        Table.TableName        := DatasetName;
        Table.SessionName      := FSessionName;
        Table.PrimaryIndexName := ORMapping.GetPrimaryIndex(DatasetName).Name;
        InternalSetIndexName(Handle, ORMapping.GetPrimaryIndex(DatasetName).Name);
        Table.Open;
    end;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalInsertSQL(
                  const Handle      : TOdbHandle;
                  const Stream      : TMemoryStream;
                  const ORMapping   : TORMapping;
                  const DatasetName : String );
var
    EditQuery   : TQuery;
begin
    if not SqlBased then
    begin
        OpenTableForObjectUpdate( Handle, ORMapping, DatasetName );

        InternalInsert(Handle, ORMapping.Fields, Stream );
    end
    else
    begin
        EditQuery   := GetHandleDataObj(Handle) as TQuery;

        { MB Sep 28, 2000 This method by first adding with a direct SQL statement,
           followed by a blob update is faster on datasets containing above
           2000 records. On 500 inserts it took 53 seconds compared to 42 using
           EditQuery.Insert and a select * with RequestLive (local interbase 5.X
           server), but with 5000 objects, this approach took 630 seconds compared
           to 1262 when using EditQuery.Insert. }
        BuildInsertStatement( EditQuery.SQL, ORMapping.Fields, DatasetName );

        EditQuery.ExecSQL;

        { add the blob }
        BuildSelectObjectQuery( EditQuery.SQL, ORMapping, DatasetName );
        EditQuery.RequestLive := True;
        EditQuery.Open;
        try
            if EditQuery.Bof and EditQuery.Eof then
                RaiseOdbException(EOdbObjectNotFound, 'TBdeOdbDriver.InternalInsertSQL');

            EditQuery.Edit;

            UpdateCurrentInDataset(EditQuery, ORMapping.Fields, Stream);
        finally
            EditQuery.Close;
        end;

        (* See comment above.
        { Insert using EditQuery.Insert is faster. See comment above }
        EditQuery.RequestLive  := True;
        EditQuery.SQL.Clear;
        EditQuery.SQL.Add( 'select * from ' + ObjectQuery.DatasetName );
        EditQuery.Open;
        try
            { add the record using the related edit query }
            InsertInDataset(TQuery(Dataset.Tag), Fields, Stream);
        finally
            EditQuery.Close;
        end;
        *)
    end;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalUpdateSQL(
                      const Handle      : TOdbHandle;
                      const Stream      : TMemoryStream;
                      const ORMapping   : TORMapping;
                      const DatasetName : String );
var
    EditQuery     : TQuery;
begin
    if not SqlBased then
    begin
        OpenTableForObjectUpdate( Handle, ORMapping, DatasetName );

        if not InternalFindExact( Handle, ORMapping.Fields, ORMapping.GetPrimaryIndex(DatasetName) ) then
            RaiseOdbException( EOdbObjectNotFound, 'TBdeOdbDriver.InternalUpdateSQL' );

        InternalUpdate(Handle, ORMapping.Fields, Stream );
    end
    else
    begin

        EditQuery   := GetHandleDataObj(Handle) as TQuery;

        BuildSelectObjectQuery( EditQuery.SQL, ORMapping, DatasetName );

        EditQuery.Open;
        try
            if EditQuery.Bof and EditQuery.Eof then
                RaiseOdbException(EOdbObjectNotFound, 'TBdeOdbDriver.InternalUpdateSQL');

            EditQuery.Edit;

            UpdateCurrentInDataset(EditQuery, ORMapping.Fields, Stream);
        finally
            EditQuery.Close;
        end;
    end;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalDeleteSQL(
                             const Handle      : TOdbHandle;
                             const ORMapping   : TORMapping;
                             const DatasetName : String );
var
    EditQuery   : TQuery;
begin
    if not SqlBased then
    begin
        OpenTableForObjectUpdate( Handle, ORMapping, DatasetName );

        if not InternalFindExact( Handle, ORMapping.Fields, ORMapping.GetPrimaryIndex(DatasetName) ) then
            RaiseOdbException( EOdbObjectNotFound, 'TBdeOdbDriver.InternalDeleteSQL' );

        InternalDelete(Handle);
    end
    else
    begin
        EditQuery   := GetHandleDataObj(Handle) as TQuery;

        BuildDeleteStatement( EditQuery.SQL, ORMapping, DatasetName );

        EditQuery.ExecSQL;
    end;
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalGetIndexName( const Handle : TOdbHandle ) : String;
var
    Table : TOdbBdeTable;
begin
    Table  := GetBdeDataset(Handle) as TOdbBdeTable;
    Result := Table.IndexName;

    if (FTableType = ttParadox) and (Result = '') then
        Result := Table.PrimaryIndexName;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalSetIndexName(
                      const Handle    : TOdbHandle;
                      const IndexName : String );
var
    Table : TOdbBdeTable;
begin
    Table := GetBdeDataset(Handle) as TOdbBdeTable;

    if (FTableType = ttParadox) and (IndexName = Table.PrimaryIndexName) then
        Table.IndexName := ''
    else
        Table.IndexName := IndexName;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalGetCurrent(
                      const Handle  : TOdbHandle;
                      const Stream  : TMemoryStream );
var
    Dataset : Db.TDataset;
    Field   : TBlobField;
begin
    Dataset := GetBdeDataset(Handle) as Db.TDataset;

    Field := (Dataset.FieldByName(CObjectFieldName) as TBlobField);

    Field.SaveToStream(Stream);
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalLockCurrent( const Handle : TOdbHandle ) : Boolean;
var
    Table : TOdbBdeTable;
    DbResult : DBIResult;
begin
    Table := GetBdeDataset(Handle) as TOdbBdeTable;

    Table.UpdateCursorPos;
    DbResult := DbiGetRecord(Table.Handle, dbiWriteLock, nil, nil);

    Result := (DbResult = DBIERR_NONE);

    if not Result and
      (DbResult <> DBIERR_ALREADYLOCKED) and
      (DbResult <> DBIERR_LOCKED) then
        DbiCheck(DbResult);
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalUnLockCurrent( const Handle : TOdbHandle );
var
    Table : TOdbBdeTable;
begin
    Table := GetBdeDataset(Handle) as TOdbBdeTable;

    Table.UpdateCursorPos;

    { MB May 21, 1998. Don't use DbiCheck to prevent raising an exception
      if the current object is not locked. This is important as the
      InternalUpdate method may attempt to unlock an object which is not
      locked. }
    DbiRelRecordLock(Table.Handle, False);
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalFindNext( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : Db.TDataset;
begin
    Dataset := GetBdeDataset(Handle);
    Dataset.Next;
    Result := not Dataset.EOF;
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalFindPrior( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : Db.TDataset;
begin
    Dataset := GetBdeDataset(Handle);
    Dataset.Prior;
    Result := not Dataset.BOF;
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalFindFirst( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : Db.TDataset;
begin
    Dataset := GetBdeDataset(Handle);
    Dataset.First;
    Result := not (Dataset.EOF and Dataset.BOF);
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalFindLast( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : Db.TDataset;
begin
    Dataset := GetBdeDataset(Handle);
    Dataset.Last;
    Result := not (Dataset.EOF and Dataset.BOF);
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalFindExact(
                    const Handle : TOdbHandle;
                    const Fields : TUnOrderedList;
                    const Index  : TOdbIndex ) : Boolean;
var
    Table : TOdbBdeTable;
begin
    Table := GetBdeDataset(Handle) as TOdbBdeTable;
    Table.SetKey;
    SetTableFieldValuesOnIndex(Fields, Index, Table);
    Result := Table.GotoKey;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalFindNearest(
                      const Handle : TOdbHandle;
                      const Fields : TUnOrderedList;
                      const Index  : TOdbIndex );
var
    Table : TOdbBdeTable;
begin
    Table := GetBdeDataset(Handle) as TOdbBdeTable;
    Table.SetKey;
    SetTableFieldValuesOnIndex(Fields, Index, Table);
    Table.GotoNearest;
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalGetBookmark( const Handle : TOdbHandle ) : Pointer;
var
    Table : TOdbBdeTable;
begin
    Table := GetBdeDataset(Handle) as TOdbBdeTable;

    Result := Table.GetBookmark;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalGotoBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
var
    Table : TOdbBdeTable;
begin
    Table := GetBdeDataset(Handle) as TOdbBdeTable;

    Table.GotoBookmark(Bookmark);
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalFreeBookmark( const Handle   : TOdbHandle;
                                              const Bookmark : Pointer );
var
    Table : TOdbBdeTable;
begin
    Table := GetBdeDataset(Handle) as TOdbBdeTable;

    Table.FreeBookmark(Bookmark);
end;


{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalGetObjectCount( const Handle : TOdbHandle ) : LongInt;
begin
    Result := (GetBdeDataset(Handle) as TOdbBdeTable).RecordCount;
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.InternalIsEmpty( const Handle : TOdbHandle ) : Boolean;
var
    Dataset : TDataset;
begin
    Dataset := GetBdeDataset(Handle);
    Result  := Dataset.BOF and Dataset.EOF;
end;

{--------------------------------------------------------------------}

procedure TBdeOdbDriver.InternalPack( const Handle : TOdbHandle );
var
    Table : TOdbBdeTable;
begin
    Table := GetBdeDataset(Handle) as TOdbBdeTable;
    if Table.TableType = ttDBase then
    begin
        if Table.Exclusive then
            DbiPackTable(FDatabase.Handle, Table.Handle, nil, nil, True)
        else
            RaiseOdbException( EBdeTableNotOpenedExclusively, '' );
    end;
    { MB Mar 10, 2001. Don't raise exception if not DBase table. This allows
      the automated tests to pass, and the "pack" operation itself can be
      argued to be optional }
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.GetBdeDataset( const Handle : TOdbHandle ) : Db.TDataset;
begin
    Result := GetHandleDataObj(Handle) as Db.TDataset;
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.GetDataAccessMethods : TDataAccessMethodSet;
begin
    { MB Mar 26, 2001. ISAM access with SQL databases should be avoided.
      The involves using TTable objects with SQL backends. This does work
      to a certain, but when testing with BDE and Interbase, the automated
      tests failed when attempting to create a second table }
    if FSqlBased then
        Result := [damSQL]
    else
        Result := [damISAM, damSQL];
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.GetSupportedFieldTypes : TOdbFieldTypes;
begin
    Result := [oftString, oftInteger, oftFloat];

    {$ifndef Ver80}
    { Delphi 1 does not define the field type ftAutoInc }
    Result := Result + [oftAutoInc];
    {$endif}
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.CheckDuplicateInsert : Boolean;
begin
    { A unique index in DBase is very confusing. It does not mean
      unique values, but rather duplicates are ignored when
      navigating, creating an undesirable behavior. For this reason,
      duplicates are checked by the TObjectFile instance. }
    Result := (FTableType = ttDBase);
end;

{--------------------------------------------------------------------}

function TBdeOdbDriver.CheckDuplicateUpdate : Boolean;
begin
    { A unique index in DBase is very confusing. It does not mean
      unique values, but rather duplicates are ignored when
      navigating, creating an undesirable behavior. For this reason,
      duplicates are checked by the TObjectFile instance. }
    Result := (FTableType = ttDBase);
end;

{--------------------------------------------------------------------------}

function TBDEOdbDriver.AutoIncSupported              : Boolean;
begin
    { MB Mar 26, 2001. BDE with Interbase tests failed with autoincrement
      field }
    if FDatabase <> nil then
    begin
        Result := (not FSqlBased) and
                  (FtableType = ttParadox);
    end
    else
        Result := True;
end;

{--------------------------------------------------------------------------}

function TBDEOdbDriver.IntegerIndexSupported         : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBDEOdbDriver.FloatIndexSupported           : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBDEOdbDriver.CaseInsensitiveIndexSupported : Boolean;
begin
    { MB Mar 26, 2001. BDE with Interbase and ISAM tests failed with
      case insensitive index }
    if FDatabase <> nil then
    begin
        Result := (not FSqlBased) and
                  (FtableType = ttParadox);
    end
    else
        Result := False;
end;

{--------------------------------------------------------------------------}

function TBDEOdbDriver.DescendingIndexSupported      : Boolean;
begin
    { MB Mar 26, 2001. BDE with Interbase and ISAM tests failed with
      case insensitive index }
    if FDatabase <> nil then
    begin
        Result := (not FSqlBased) and
                  (FtableType = ttParadox);
    end
    else
        Result := False;
end;

{--------------------------------------------------------------------------}

function TBDEOdbDriver.TransactionsSupported         : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBDEOdbDriver.FileLockingSupported          : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBDEOdbDriver.ObjectLockingSupported        : Boolean;
begin
    Result := True;
end;

{--------------------------------------------------------------------------}

function TBDEOdbDriver.SingleIndexOnly               : Boolean;
begin
    Result := False;
end;

{--------------------------------------------------------------------}

{$ifdef __Testing}

procedure TParadoxBDEOdbDriverTest.GetParams(
                    var DriverClass                   : TOdbDriverClass;
                    var DatabaseName                  : String;
                    var ServerName                    : String;
                    var UserName                      : String;
                    var Password                      : String
                  );
begin
    DriverClass  := TBDEOdbDriver;
    DatabaseName := Self.TestDbPath;
    ServerName   := '';
    UserName     := '';
    Password     := '';
end;

procedure TInterbaseBDEOdbDriverTest.GetParams(
                    var DriverClass                   : TOdbDriverClass;
                    var DatabaseName                  : String;
                    var ServerName                    : String;
                    var UserName                      : String;
                    var Password                      : String
                  );
begin
    DriverClass  := TBDEOdbDriver;
    DatabaseName := 'InterbaseOdb';
    ServerName   := '';
    UserName     := 'sysdba';
    Password     := 'masterkey';
end;

{$endif}

{--------------------------------------------------------------------}

begin
    RegisterOdbDriver(TBdeOdbDriver, 'BDE');
end.
