{---------------------------------------------------------------------------
|
|  Module: OdbMisc.Pas
|
|  Description: Miscellanous routines, contants and types.
|
|  History: Jan 17, 1998. MB first version
|           Apr 04, 1998. MB add TimeToSortString
|                                DateToSortString
|                                DateTimeToSortString
|           May 17, 1998. MB add WordCount, ExtractWord, WordPosition
|           Feb   , 2001. MB Create unit tests for all routines
|           Apr 10, 2001. MB add debug window hooks
|
|---------------------------------------------------------------------------}
unit OdbMisc;

{$I Odb.Inc}

interface

uses
    {$ifdef __Testing}
    AutoTest, { for TSpiderOdbTestcase }
    {$endif}

    Classes,  { for TStrings, TFileStream }
    ObjList,  { for TUnOrderedList }
    SysUtils, { for Exception, FileGetAttr }
    Graphics; { for TBitmap }

type

TOdbCharSet = set of Char;

TExceptionClass = class of Exception;

TOdbExceptionProc = procedure( const ExceptionClass : TExceptionClass;
                               const Msg            : String );

{ debug output types }
TDebugOutputStrProc = procedure( const Msg   : String);
TDebugOutputIntProc = procedure( const Msg   : String;
                                 const Value : Integer);

var

{ debug output procedures. Maybe changed to replace defaults }
DebugOutputStr : TDebugOutputStrProc;
DebugOutputInt : TDebugOutputIntProc;

{ send

{--------------------------------------------------------------------}

{ Send a string list to the debug window. Use to check SQL statements when
  debugging }
procedure DebugOutputStrings( const SQL : TStrings );

{ add a back slash to a path, as appropriate }
function AddBackSlashToPath(const Path: String): String;

{ remove a back slash from a path }
function RemoveBackSlashFromPath(const Path: String): String;

{ delete the file(s) specified by FileName. FileName may contain wild
  cards. Read-only and system files will NOT be deleted. }
procedure DeleteFiles( const FileName : String );

{ Find all files which match FileName (which may contain wild cards) and
  match the attributes specified in Attributes. Attributes is a bit map of
  the following constants in the SysUtils unit,

  faReadOnly	$01	Read-only files
  faHidden	$02	Hidden files
  faSysFile	$04	System files
  faVolumeID	$08	Volume ID files
  faDirectory	$10	Directory files
  faArchive	$20	Archive files
  faAnyFile	$3F	Any file

  FileNameList is a list containing TStringCombo's.

  For example, FindFiles( 'c:\windows\system\*.dll',
                          faArchive or faReadOnly,
                          FileNameList );
}
procedure FindFiles( const FileName     : String;
                     const Attributes   : Word;
                     const FileNameList : TUnOrderedList );


{ Returns an object file name for local databases. If the file
  name contains path information, the file name is returned unchanged. If the
  file name does not contain path information, the file name is appended to the
  Path. }
function GetLocalObjectFileName( const Path      : String;
                                 const FileName  : String;
                                 const Extension : String ) : String;

{ Raises an exception. The the default exception procedure uses the
  class name of the exception followed by an optional message. The
  default behavior may be modified with SetOdbExceptionProc. }
procedure RaiseOdbException( const ExceptionClass : TExceptionClass;
                             const Msg            : String );

{ Sets the exception procedure. This may be used to translate exception
  names for multilanguage support. }
procedure SetOdbExceptionProc( const ExceptionProc : TOdbExceptionProc );

{ returns True if all the bit fields in Flags are set in Value. }
function AreFlagsSet(const Value, Flags : LongInt) : Boolean;

{ Set the bit fields contained in Flags in Value. }
procedure SetLongFlags(var Value : LongInt; const Flags : LongInt);

{ Clear the bit fields contained in Flags from Value. }
procedure ClearLongFlags(var Value : LongInt; const Flags : LongInt);

{ Set the bit fields contained in Flags in Value. }
procedure SetWordFlags(var Value : Word; const Flags : Word);

{ Clear the bit fields contained in Flags from Value. }
procedure ClearWordFlags(var Value : Word; const Flags : Word);

{ Process a message on the application's message queue. On return
  Quit is True, if the WM_QUIT message has been encountered. }
procedure ProcessMessage( var Quit : Boolean);

{ Wait for MilliSeconds while yielding to other applications. }
procedure Delay( const MilliSeconds : LongInt );

{ Truncate Str to a length of MaxLen characters. }
function TruncateStr( const Str    : String;
                      const MaxLen : Integer ) : String;

{ Create a string of a specified length given a character }
function CreateStr( const Ch  : Char;
                    const Len : Integer ) : String;

{ Returns Str padded with spaces on the left of the string. If Len is
  smaller than the length of Str, the string will be truncated. }
function LeftPad( const Str : String;
                  const Ch  : Char;
                  const Len : Integer): String;

{ Returns Str padded with spaces on the left of the string. If Len is
  smaller than the length of Str, the string will be truncated. }
function RightPad( const Str : String;
                   const Ch  : Char;
                   const Len : Integer ) : String;

{ returns date/time values as sortable strings. }
function TimeToSortString( const Time : TDateTime ) : String;
function DateToSortString( const Date : TDateTime ) : String;
function DateTimeToSortString( const DateTime : TDateTime ) : String;

{ converts a time to a time string. The format of the time is defined
  by Mask, ie, 'hh:nn:ss'. }
function TimeToTimeStr( const Time : TDateTime;
                        const Mask : String ) : String;

{ converts a time string to a time. The format of the time is defined
  by Mask, ie, 'hh:nn:ss'. }
function TimeStrToTime( const Str   : String;
                        const Mask  : String ) : TDateTime;

{ converts a date to a date string. The format of the date is defined
  by Mask, ie, 'yyyy/mm/dd'. The date separator (ie, / ) must be one of
  ['.', '-','/'] }
function DateToDateStr( const Date  : TDateTime;
                        const Mask  : String ) : String;

{ converts a date string to a date. The format of the date is defined
  by Mask, ie, 'yyyy/mm/dd'. }
function DateStrToDate( const Str   : String;
                        const Mask  : String ) : TDateTime;

{ returns the number of words in Str which are delimited by any
  character in Delimiters. }
function WordCount( const Str        : String;
                    const Delimiters : TOdbCharSet ): Integer;

{ extract the N'th word in Str given a set of delimiters. }
function ExtractWord( const N          : Integer;
                      const Str        : String;
                      const Delimiters : TOdbCharSet ) : String;

{ returns the position of the N'th word in Str given a set of delimiters. }
function WordPosition( const N          : Integer;
                       const Str        : String;
                       const Delimiters : TOdbCharSet ): Integer;

{ searches for Parameter in the command line and returns the value of the
  parameter. The must NOT be a space between the paramater and the value.
  The value must not contain any spaces. Returns False if the command line
  parameter is not specified. The search is NOT case sensitive.

  For example, with a  '/f=500' on the command line, calling
  ExtractCommandLineValue('/f=', Value) will return 500 in Value. }
function ExtractCommandLineValue( const Parameter : String;
                                  var   Value     : String ) : Boolean;

{ returns a bitmap found in the resources of the executable file. Returns
  nil if unable to create or load bitmap. }
function LoadBitmapFromExeRes( const ResName : String ) : Graphics.TBitmap;


{$ifdef __Testing}

type

TOdbMiscUnitTest = class( TSpiderOdbTestCase )
published
    procedure TestExtractCommandLineValue;
    procedure TestAddBackSlashToPath;
    procedure TestRemoveBackSlashFromPath;
    procedure TestDeleteFiles;
    procedure TestFindFiles;
    procedure TestGetLocalObjectFileName;
    procedure TestRaiseOdbException;
    procedure TestSetOdbExceptionProc;
    procedure TestAreFlagsSet;
    procedure TestSetLongFlags;
    procedure TestClearLongFlags;
    procedure TestSetWordFlags;
    procedure TestClearWordFlags;
    procedure TestProcessMessage;
    procedure TestDelay;

    procedure TestTruncateStr;
    procedure TestCreateStr;
    procedure TestLeftPad;
    procedure TestRightPad;

    procedure TestTimeToSortString;
    procedure TestDateToSortString;
    procedure TestDateTimeToSortString;

    procedure TestTimeConversion;

    procedure TestDateConversion;

    procedure TestWordCount;
    procedure TestExtractWord;
    procedure TestWordPosition;
    procedure TestLoadBitmapFromExeRes;

end;

{$endif}

{--------------------------------------------------------------------}

implementation

uses
    {$ifdef __DebugWindow}
    DBugIntf, { for SendDebug }
    {$endif}

    Messages, { for WM_QUIT }
    {$ifdef Win32}
    Windows,  { for GetTickCount, TMsg }
    {$else}
    WinProcs, { for GetTickCount }
    WinTypes, { for TMsg }
    {$endif}
    ObjBuckt; { for TStringCombo }

var

OdbExceptionProc : TOdbExceptionProc;

{--------------------------------------------------------------------------}

procedure DebugOutputStrings( const SQL : TStrings );
var
    LineNo : Integer;
begin
    for LineNo := 0 to (SQL.Count-1) do
        DebugOutputStr( SQL[LineNo] );
end;

{--------------------------------------------------------------------------}

function AddBackSlashToPath(const Path: String): String;
begin
    if Path = '' then
    begin
        Result := '\';
        Exit;
    end;

    Result := Path;

    { verify Path does not end with :, before adding the back slash. ie, to
      prevent c: becoming c:\, which is not the same directory. }
    if (Path[Length(Path)] <> '\') and
       (Path[Length(Path)] <> ':') then
        Result := Path + '\';
end;

{--------------------------------------------------------------------------}

function RemoveBackSlashFromPath(const Path : String): String;
begin
    Result := Path;

    if Path = '' then
        Exit;

    if Path[Length(Path)] = '\' then
        Result := Copy(Path,1,Length(Path)-1);
end;

{----------------------------------------------------------------------------}

{ Add an extension to a file, if the file does NOT have an extension. }
function AddFileExtension( const FileName  : String;
                           const Extension : String ) : String;
begin
    Result := FileName;
    if (ExtractFileExt(FileName) <> '') or
       (Extension = '') then
        Exit;

    if Extension[1] = '.' then
        Result := FileName + Extension
    else
        Result := FileName + '.' + Extension;
end;

{----------------------------------------------------------------------------}

function GetLocalObjectFileName( const Path      : String;
                                 const FileName  : String;
                                 const Extension : String ) : String;
begin
    if ExtractFilePath(FileName) <> '' then
        Result := FileName
    else
        Result := AddBackSlashToPath(Path) + FileName;

    Result := AddFileExtension(Result, Extension);
end;

{----------------------------------------------------------------------------}

procedure DefaultRaiseException( const ExceptionClass : TExceptionClass;
                                 const Msg            : String ); far;
var
    FullMsg : String;
begin
    FullMsg := ExceptionClass.ClassName;
    if Msg <> '' then
        FullMsg := FullMsg + '.' + #13 + Msg;
    raise ExceptionClass.Create( FullMsg );
end;

{----------------------------------------------------------------------------}

procedure RaiseOdbException( const ExceptionClass : TExceptionClass;
                             const Msg            : String );
begin
    OdbExceptionProc( ExceptionClass, Msg );
end;

{----------------------------------------------------------------------------}

procedure SetOdbExceptionProc( const ExceptionProc : TOdbExceptionProc );
begin
    OdbExceptionProc := ExceptionProc;
end;

{----------------------------------------------------------------------------}

function AreFlagsSet(const Value, Flags : LongInt) : Boolean;
begin
    Result := (Flags = Flags and Value);
end;

{------------------------------------------------------------------------}

procedure SetLongFlags(var Value : LongInt; const Flags : LongInt);
begin
    Value := Value or Flags;
end;

{------------------------------------------------------------------------}

procedure ClearLongFlags(var Value : LongInt; const Flags : LongInt);
begin
    Value := Value and (not Flags);
end;

{----------------------------------------------------------------------------}

procedure SetWordFlags(var Value : Word; const Flags : Word);
begin
    Value := Value or Flags;
end;

{------------------------------------------------------------------------}

procedure ClearWordFlags(var Value : Word; const Flags : Word);
begin
    Value := Value and (not Flags);
end;

{--------------------------------------------------------------------------}

procedure ProcessMessage( var Quit : Boolean);
var
   Msg            : TMsg;

begin
    Quit := False;

    { Application.ProcessMessages is not used since in WinCrt
      applications, the Application variable is not initialized. }
    if PeekMessage( Msg, 0, 0, 0, PM_REMOVE ) then
    begin
        if Msg.Message = WM_QUIT then
        begin
            PostQuitMessage(Msg.wParam);
            Quit := True;
        end
        else
        begin
            TranslateMessage(Msg);
            DispatchMessage(Msg);
        end;
    end;
end;

{------------------------------------------------------------------------}

procedure Delay( const MilliSeconds : LongInt );
var
   FirstTickCount : LongInt;
   Quit           : Boolean;
begin
    FirstTickCount := GetTickCount;
    repeat
        ProcessMessage(Quit);

        if Quit then
            break;

    {$ifdef __Delphi4AndAbove}
    until (GetTickCount-DWORD(FirstTickCount)) >= DWORD(MilliSeconds);
    {$else}
    until (GetTickCount-FirstTickCount) >= MilliSeconds;
    {$endif}
end;

{------------------------------------------------------------------------}

function TruncateStr( const Str    : String;
                      const MaxLen : Integer ) : String;
begin
    Result := Str;

    System.Delete(Result, MaxLen+1, Length(Result));
end;

{------------------------------------------------------------------------}

function CreateStr( const Ch  : Char;
                    const Len : Integer ) : String;
var
   CharNo  : Integer;
begin
   Result := '';
   for CharNo := 1 to Len do
       Result := Result + Ch;
end;

{------------------------------------------------------------------------}

function LeftPad( const Str : String;
                  const Ch  : Char;
                  const Len : Integer ) : String;
begin
   if Length(Str) < Len then
       Result := CreateStr(' ', Len - Length(Str)) + Str
   else
       Result := Copy(Str, Length(Str)-Len + 1, Len);
end;

{--------------------------------------------------------------------------}

function RightPad( const Str : String;
                   const Ch  : Char;
                   const Len : Integer ) : String;
begin
   if Length(Str) < Len then
       Result := Str + CreateStr(' ', Len - Length(Str))
   else
       Result := Copy(Str, 1, Len);
end;

{------------------------------------------------------------------------}

function TimeToSortString( const Time : TDateTime ) : String;
begin
    Result := TimeToTimeStr( Time, 'hhnnss');
end;

{------------------------------------------------------------------------}

function DateToSortString( const Date : TDateTime ) : String;
begin
    Result := DateToDateStr(Date, 'yyyymmdd');
end;

{------------------------------------------------------------------------}

function DateTimeToSortString( const DateTime : TDateTime ) : String;
begin
    Result := DateToSortString(DateTime) +
              TimeToSortString(DateTime);
end;

{------------------------------------------------------------------------}

function GetDateSeparator( const Mask : String ) : Char;
begin
    if Pos('.', Mask) <> 0 then
        Result := '.'
    else if Pos('-', Mask) <> 0 then
        Result := '-'
    else
        Result := '/';
end;


{------------------------------------------------------------------------}

function DateToDateStr( const Date  : TDateTime;
                        const Mask  : String ) : String;
var
    SavedStr       : String;
    SavedSeparator : Char;
begin
    SavedStr       := ShortDateFormat;
    SavedSeparator := DateSeparator;
    try
        ShortDateFormat := Mask;
        DateSeparator   := GetDateSeparator(Mask);
        Result := DateToStr(Date);
    finally
        DateSeparator   := SavedSeparator;
        ShortDateFormat := SavedStr;
    end;
end;

{------------------------------------------------------------------------}

function DateStrToDate( const Str   : String;
                        const Mask  : String ) : TDateTime;
var
    SavedStr       : String;
    SavedSeparator : Char;
begin
    SavedStr       := ShortDateFormat;
    SavedSeparator := DateSeparator;
    try
        ShortDateFormat := Mask;
        DateSeparator   := GetDateSeparator(Mask);
        Result := StrToDate(Str);
    finally
        DateSeparator   := SavedSeparator;
        ShortDateFormat := SavedStr;
    end;
end;

{------------------------------------------------------------------------}

function TimeToTimeStr( const Time : TDateTime;
                        const Mask : String ) : String;
var
    SavedStr : String;
begin
    { use LongTimeFormat to include seconds. }
    SavedStr := LongTimeFormat;
    try
        LongTimeFormat := Mask;
        Result := TimeToStr(Time);
    finally
        LongTimeFormat := SavedStr;
    end;
end;

{------------------------------------------------------------------------}

function TimeStrToTime( const Str   : String;
                        const Mask  : String ) : TDateTime;
var
    SavedStr : String;
begin
    SavedStr := LongTimeFormat;
    try
        LongTimeFormat := Mask;
        Result := StrToTime(Str);
    finally
        LongTimeFormat := SavedStr;
    end;
end;

{------------------------------------------------------------------------}

procedure DeleteFiles( const FileName : String );
var
    FileNameList : TUnOrderedList;
begin
    FileNameList := TUnOrderedList.Create(TStringCombo, CompareStringComboNoCase);
    try
        FindFiles(FileName, faArchive, FileNameList);

        if FileNameList.GotoFirst then
        repeat
            SysUtils.DeleteFile((FileNameList.CurrentObj as TStringCombo).Str);
        until not FileNameList.GotoNext;

    finally
        FileNameList.Free;
    end;
end;

{------------------------------------------------------------------------}

procedure FindFiles( const FileName     : String;
                     const Attributes   : Word;
                     const FileNameList : TUnOrderedList );

var
    SearchRec  : TSearchRec;
    FindResult : Integer;
    Include    : Boolean;
begin
    { MB Feb 06, 2001. Clear FileNameList }
    FileNameList.Clear;
    
    FindResult := FindFirst(FileName, Attributes, SearchRec);
    try
        while FindResult = 0 do
        begin                       
            { files without any attributes are included automatically. }
            Include := (SearchRec.Attr = 0) or
                       ((SearchRec.Attr and Attributes) <> 0);

            { if the file is the '.' or '..' directories, ignore }
            if ((SearchRec.Attr and faDirectory and Attributes) <> 0) and
                (SearchRec.Name[1] = '.') then
                 Include := False;

            if Include then
                FileNameList.Insert(
                       TStringCombo.Create( AddBackSlashToPath(
                                             ExtractFilePath(FileName)) +
                                            SearchRec.Name, nil));

            FindResult := FindNext(SearchRec);
        end;
    finally
        { MB Feb 06, 2000. Add FindClose }
        SysUtils.FindClose(SearchRec);
    end;

end;

{------------------------------------------------------------------------}

function WordCount( const Str        : String;
                    const Delimiters : TOdbCharSet ): Integer;
var
  StrLen   : LongInt;
  Position : LongInt;
begin
    Result   := 0;
    Position := 1;
    StrLen   := Length(Str);
    while Position <= StrLen do
    begin
        while (Position <= StrLen) and (Str[Position] in Delimiters) do
            Inc(Position);

        if Position <= StrLen then
            Inc(Result);

        while (Position <= StrLen) and not (Str[Position] in Delimiters) do
            Inc(Position);
    end;
end;

{------------------------------------------------------------------------}

function WordPosition( const N          : Integer;
                       const Str        : String;
                       const Delimiters : TOdbCharSet ): Integer;
var
  Count    : LongInt;
  Index : LongInt;
begin
    Count    := 0;
    Index    := 1;
    Result   := 0;
    while (Index <= Length(Str)) and (Count <> N) do
    begin
        while (Index <= Length(Str)) and (Str[Index] in Delimiters) do
            Inc(Index);

        if Index <= Length(Str) then
            Inc(Count);

        if Count <> N then
            while (Index <= Length(Str)) and not (Str[Index] in Delimiters) do
                Inc(Index)
        else
            Result := Index;
    end;
end;

{------------------------------------------------------------------------}

function ExtractWord( const N          : Integer;
                      const Str        : String;
                      const Delimiters : TOdbCharSet ) : String;
var
    Position : LongInt;
begin
    Result   := '';

    Position := WordPosition(N, Str, Delimiters);
    if Position = 0 then
        Exit;

    while (Position <= Length(Str)) and not (Str[Position] in Delimiters) do
    begin
        Result := Result + Str[Position];
        Inc(Position);
    end;
end;

{------------------------------------------------------------------------}

type

TParamStrFunc   = function(Index: Integer): String;
TParamCountFunc = function : Integer;

function ExtractCommandLineValueInternal( const Parameter      : String;
                                          var   Value          : String;
                                          const ParamCountFunc : TParamCountFunc;
                                          const ParamStrFunc   : TParamStrFunc
                                         ) : Boolean;
var
    ParameterNo : Integer;
    Position    : Integer;
begin
    Result  := False;
    Value   := '';

    for ParameterNo := 1 to ParamCountFunc do
    begin
        Position := Pos( AnsiUpperCase(Parameter),
                         AnsiUpperCase(ParamStrFunc(ParameterNo)));
        if Position > 0 then
        begin
            Value  := Copy( ParamStrFunc(ParameterNo),
                            Position + Length(Parameter),
                            Length(ParamStrFunc(ParameterNo)));


            Result := True;
            break;
        end;
    end;
end;

{---------------------------------------------------------------------------}

function ExtractCommandLineValue( const Parameter      : String;
                                  var   Value          : String ) : Boolean;
begin
    Result := ExtractCommandLineValueInternal( Parameter, Value,
                                               System.ParamCount,
                                               System.ParamStr );
end;

{---------------------------------------------------------------------------}

function LoadBitmapFromExeRes( const ResName : String ) : Graphics.TBitmap;
var
    Handle  : THandle;
    ZString : array[1..128] of Char;
begin
    Handle := LoadBitmap( HInstance,
                          StrPLCopy(@ZString[1], ResName, SizeOf(ZString)));

    if Handle = 0 then
        Result := nil
    else
    begin
        Result := Graphics.TBitmap.Create;
        Result.Handle := Handle;
    end;
end;

{$ifdef __DebugWindow}
procedure OutputStr( const Msg   : String);
begin
    DBugIntf.SendDebug(Msg);
end;

procedure OutputInt( const Msg   : String;
                     const Value : Integer);
begin
    DBugIntf.SendInteger( Msg, Value );
end;
{$else}
procedure OutputStr( const Msg   : String);
begin
end;

procedure OutputInt( const Msg   : String;
                     const Value : Integer);
begin
end;
{$endif}

{--------------------------------------------------------------------------}

{$ifdef __Testing}

{$R Bitmap.Res}

function ParamCountTest : Integer;
begin
    Result := 10;
end;

function ParamStrTest( Index : Integer ) : String;
begin
    if (Index mod 2) = 0 then
        Result := '/' + IntToStr(Index) + '=' + IntToStr(Index*2)
    else
        Result := '-' + IntToStr(Index*Index);

end;

procedure TOdbMiscUnitTest.TestExtractCommandLineValue;
var
    Value : String;
begin
    Check( not ExtractCommandLineValueInternal( '/f=500', Value,
                                                ParamCountTest,
                                                ParamStrTest ) );

    { first parameter }
    Check( ExtractCommandLineValueInternal( '-1', Value,
                                            ParamCountTest,
                                            ParamStrTest ) and
           (Value = '') );

    Check( ExtractCommandLineValueInternal( '/6=', Value,
                                            ParamCountTest,
                                            ParamStrTest ) and
           (Value = '12') );

    { last parameter }
    Check( ExtractCommandLineValueInternal( '/10=', Value,
                                            ParamCountTest,
                                            ParamStrTest ) and
           (Value = '20') );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestAddBackSlashToPath;
begin
    Check( AddBackSlashToPath('') = '\' );

    { special case where adding a back slash changes the directory }
    Check( AddBackSlashToPath('c:') = 'c:' );

    Check( AddBackSlashToPath('c:\') = 'c:\' );

    Check( AddBackSlashToPath('c:temp') = 'c:temp\' );

    Check( AddBackSlashToPath('c:temp\') = 'c:temp\' );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestRemoveBackSlashFromPath;
begin
    Check( RemoveBackSlashFromPath('') = '' );

    Check( RemoveBackSlashFromPath('\') = '' );

    Check( RemoveBackSlashFromPath('c:') = 'c:' );

    Check( RemoveBackSlashFromPath('c:\') = 'c:' );

    Check( RemoveBackSlashFromPath('c:temp') = 'c:temp' );

    Check( RemoveBackSlashFromPath('c:temp\') = 'c:temp' );
end;

{------------------------------------------------------------------------}

procedure CreateFile( const FileName : String;
                      const ReadOnly : Boolean );
var
    Stream : TFileStream;
    Data   : Byte;

begin
    Stream := TFileStream.Create(FileName, fmCreate);
    try
        { add a byte because SysUtils.DeleteFile does not work with
          empty files. }
        Data := 66;
        Stream.Write(Data, 1);
    finally
        Stream.Free;
    end;

    if ReadOnly then
        FileSetAttr( FileName, FileGetAttr(FileName) or faReadOnly);
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestDeleteFiles;
begin
    CreateFile( TestDbPath + 'File1', False );
    CreateFile( TestDbPath + 'File2', False );
    CreateFile( TestDbPath + 'File3', True );

    DeleteFiles( TestDbPath + 'File*' );

    Check( not FileExists( TestDbPath + 'File1') ) ;
    Check( not FileExists( TestDbPath + 'File2') ) ;
    Check( FileExists( TestDbPath + 'File3') ) ;

    { remove the readonly attribute and delete file 3 }
    FileSetAttr( TestDbPath + 'File3', FileGetAttr(TestDbPath + 'File3') and not faReadOnly);
    DeleteFiles( TestDbPath + 'File*' );
    Check( not FileExists( TestDbPath + 'File3') ) ;
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestFindFiles;
var
    FileNames   : TUnOrderedList;
    StringCombo : TStringCombo;
begin
    CreateFile( TestDbPath + 'File1', False );
    CreateFile( TestDbPath + 'File2', False );
    CreateFile( TestDbPath + 'File3', True );
    CreateFile( TestDbPath + 'Frame1', False );
    FileNames   := TUnOrderedList.Create(TStringCombo, CompareStringComboNoCase);
    StringCombo := TStringCombo.Create('', nil);
    try
        { most files will have the faArchive bit set, including the files
          just created.  }
        FindFiles( TestDbPath + 'File*',
                    faArchive,
                    FileNames );

        Check( FileNames.Size = 3 );

        StringCombo.Str := TestDbPath + 'File1';
        Check( FileNames.Search( StringCombo, 1 ) );

        StringCombo.Str := TestDbPath + 'File2';
        Check( FileNames.Search( StringCombo, 1 ) );

        StringCombo.Str := TestDbPath + 'File3';
        Check( FileNames.Search( StringCombo, 1 ) );

        { find only readonly files. Also add another attribute to test that
          the attribute comparison matches any of the attributes }
        FindFiles( TestDbPath + '*.*',
                    faReadOnly or faHidden,
                    FileNames );

        Check( FileNames.Size = 1 );

        StringCombo.Str := TestDbPath + 'File3';
        Check( FileNames.Search( StringCombo, 1 ) );


    finally
        StringCombo.Free;
        FileNames.Free;
        FileSetAttr( TestDbPath + 'File3', FileGetAttr(TestDbPath + 'File3') and not faReadOnly);
        DeleteFiles( TestDbPath + 'F*' );
    end;
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestGetLocalObjectFileName;
begin
    { check file that already has path }
    Check( GetLocalObjectFileName( TestDbPath,
                                   TestDbPath + 'File1.Dat',
                                   'Dat' ) = TestDbPath + 'File1.Dat');

    Check( GetLocalObjectFileName( TestDbPath,
                                   'File1',
                                   'Dat' ) = TestDbPath + 'File1.Dat');


    Check( GetLocalObjectFileName( TestDbPath,
                                   'File1',
                                   '.Dat' ) = TestDbPath + 'File1.Dat');
end;

{------------------------------------------------------------------------}

type

ETestException = class(Exception);

procedure TOdbMiscUnitTest.TestRaiseOdbException;
begin
    try
        RaiseOdbException( ETestException, 'Test message');
    except
        on E : ETestException do
        begin
            Check( E.Message = E.ClassName + '.' + #13 + 'Test message');
        end;

        on E : Exception do
        begin
            raise;
        end;
    end;
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestSetOdbExceptionProc;
begin
    SetOdbExceptionProc(nil);
    Check( @OdbExceptionProc = nil );

    SetOdbExceptionProc(DefaultRaiseException);
    Check( @OdbExceptionProc = @DefaultRaiseException );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestAreFlagsSet;
begin
    Check( not AreFlagsSet($0000, $0001) );
    Check( not AreFlagsSet($0400, $0001) );
    Check( AreFlagsSet($0401, $0001) );
    Check( AreFlagsSet($0001, $0001) );
    Check( AreFlagsSet($8401, $0401) );
    Check( not AreFlagsSet($8201, $0401) );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestSetLongFlags;
var
    Value : LongInt;
begin
    Value := $41001008;
    SetLongFlags(Value, $00000000);
    Check( Value = $41001008 );

    SetLongFlags(Value, $00400001);
    Check( Value = $41401009 );

    SetLongFlags(Value, $10400001);
    Check( Value = $51401009 );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestClearLongFlags;
var
    Value : LongInt;
begin
    Value := $51401009;
    ClearLongFlags(Value, $10400001);
    Check( Value = $41001008 );

    ClearLongFlags(Value, $41400001);
    Check( Value = $00001008 );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestSetWordFlags;
var
    Value : Word;
begin
    Value := $1008;
    SetWordFlags(Value, $0000);
    Check( Value = $1008 );

    SetWordFlags(Value, $0001);
    Check( Value = $1009 );

    SetWordFlags(Value, $0001);
    Check( Value = $1009 );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestClearWordFlags;
var
    Value : Word;
begin
    Value := $1009;
    ClearWordFlags(Value, $0001);
    Check( Value = $1008 );

    ClearWordFlags(Value, $1001);
    Check( Value = $0008 );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestProcessMessage;
var
    Quit : Boolean;
begin
    { this not just a coverage test }
    ProcessMessage(Quit);
    Check( not Quit );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestDelay;
var
    TickCount : LongInt;
begin
    TickCount := Windows.GetTickCount;
    OdbMisc.Delay(124);
    Check( (LongInt(Windows.GetTickCount) - TickCount) >= 124 );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestTruncateStr;
begin
    Check( TruncateStr( '12345', 4 ) = '1234' );
    Check( TruncateStr( '12345', 5 ) = '12345' );
    Check( TruncateStr( '12345', 6 ) = '12345' );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestCreateStr;
begin
    Check( CreateStr( '1', 0 ) = '' );
    Check( CreateStr( '1', 1 ) = '1' );
    Check( CreateStr( '1', 5 ) = '11111' );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestLeftPad;
begin
    Check( LeftPad( '123', ' ', 0 ) = '' );
    Check( LeftPad( '123', ' ', 1 ) = '3' );
    Check( LeftPad( '123', ' ', 2 ) = '23' );
    Check( LeftPad( '123', ' ', 3 ) = '123' );
    Check( LeftPad( '123', ' ', 4 ) = ' 123' );
    Check( LeftPad( '123', ' ', 5 ) = '  123' );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestRightPad;
begin
    Check( RightPad( '123', ' ', 0 ) = '' );
    Check( RightPad( '123', ' ', 1 ) = '1' );
    Check( RightPad( '123', ' ', 2 ) = '12' );
    Check( RightPad( '123', ' ', 3 ) = '123' );
    Check( RightPad( '123', ' ', 4 ) = '123 ' );
    Check( RightPad( '123', ' ', 5 ) = '123  ' );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestTimeToSortString;
var
    Time3 : String;
    Time2 : String;
    Time1 : String;
begin
    Time3 := TimeToSortString(TimeStrToTime('23:01', 'hh:nn'));
    Time2 := TimeToSortString(TimeStrToTime('09:12:01', 'hh:nn:ss'));
    Time1 := TimeToSortString(TimeStrToTime('09:12', 'hh:nn'));

    Check( Time3 > Time2 );
    Check( Time2 > Time1 );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestDateToSortString;
var
    Date3 : String;
    Date2 : String;
    Date1 : String;
begin
    Date3 := DateToSortString(DateStrToDate('2001-12-01', 'yyyy-mm-dd'));
    Date2 := DateToSortString(DateStrToDate('1997.01.13', 'yyyy.mm.dd'));
    Date1 := DateToSortString(DateStrToDate('01/12/1997', 'mm/dd/yyyy'));

    Check( Date3 > Date2 );
    Check( Date2 > Date1 );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestDateTimeToSortString;
var
    DateTime3 : String;
    DateTime2 : String;
    DateTime1 : String;
begin
    DateTime3 := DateTimeToSortString( DateStrToDate('2001-12-01', 'yyyy-mm-dd') +
                                       TimeStrToTime('23:01:00', 'hh:nn:ss') );
    DateTime2 := DateTimeToSortString( DateStrToDate('2001-12-01', 'yyyy-mm-dd') +
                                       TimeStrToTime('23:00:59', 'hh:nn:ss') );
    DateTime1 := DateTimeToSortString( DateStrToDate('2001-06-01', 'yyyy-mm-dd') +
                                       TimeStrToTime('23:01:00', 'hh:nn:ss') );

    Check( DateTime3 > DateTime2 );
    Check( DateTime2 > DateTime1 );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestTimeConversion;
begin
    Check( TimeToTimeStr( TimeStrToTime('03:58', 'hh:nn') , 'hh:nn' ) = '03:58' );
    Check( TimeToTimeStr( TimeStrToTime('23:58', 'hh:nn') , 'hh:nn' ) = '23:58' );
    Check( TimeToTimeStr( TimeStrToTime('03:58:43', 'hh:nn:ss') , 'hh:nn:ss' ) = '03:58:43' );
    Check( TimeToTimeStr( TimeStrToTime('23:58:43', 'hh:nn:ss') , 'hh:nn:ss' ) = '23:58:43' );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestDateConversion;
var
    Raised : Boolean;
begin
    Check( DateToDateStr( DateStrToDate('2000.01.01', 'yyyy.mm.dd') , 'yyyy.mm.dd' ) = '2000.01.01' );
    Check( DateToDateStr( DateStrToDate('1999-12-31', 'yyyy-mm-dd') , 'yyyy-mm-dd' ) = '1999-12-31' );
    Check( DateToDateStr( DateStrToDate('99/03/30', 'yy/mm/dd') , 'yy/mm/dd' ) = '99/03/30' );

    Raised := False;
    try
        { try with invalid separator }
        Check( DateToDateStr( DateStrToDate('99:03:30', 'yy:mm:dd') , 'yy:mm:dd' ) = '99:03:30' );
    except
        on E:Exception do
            Raised := True;
    end;

    if not Raised then
        raise Exception.Create('Exception should have been raised in TestDateConversion');
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestWordCount;
begin
    Check( WordCount( '', ['%', '*']) = 0 );
    Check( WordCount( '12; 454- 23 $', ['%', '*']) = 1 );
    Check( WordCount( '12; 454- 23 $', ['2', '*']) = 3 );
    Check( WordCount( '12; 454- 23 $', ['2', ' ']) = 5 );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestExtractWord;
begin
    Check( ExtractWord( 1, '', ['%', '*']) = '');
    Check( ExtractWord( 2, '', ['%', '*']) = '');

    Check( ExtractWord( 1, '12; 454- 23 $', ['%', '*']) = '12; 454- 23 $' );
    Check( ExtractWord( 0, '12; 454- 23 $', ['%', '*']) = '' );
    Check( ExtractWord( 2, '12; 454- 23 $', ['%', '*']) = '' );

    Check( ExtractWord( 0, '12; 454- 23 $', ['2', ' ']) = '' );
    Check( ExtractWord( 1, '12; 454- 23 $', ['2', ' ']) = '1' );
    Check( ExtractWord( 2, '12; 454- 23 $', ['2', ' ']) = ';' );
    Check( ExtractWord( 3, '12; 454- 23 $', ['2', ' ']) = '454-' );
    Check( ExtractWord( 4, '12; 454- 23 $', ['2', ' ']) = '3' );
    Check( ExtractWord( 5, '12; 454- 23 $', ['2', ' ']) = '$' );
    Check( ExtractWord( 6, '12; 454- 23 $', ['2', ' ']) = '' );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestWordPosition;
begin
    Check( WordPosition( 1, '', ['%', '*']) = 0);
    Check( WordPosition( 2, '', ['%', '*']) = 0);

    Check( WordPosition( 1, '12; 454- 23 $', ['%', '*']) = 1 );
    Check( WordPosition( 0, '12; 454- 23 $', ['%', '*']) = 0 );
    Check( WordPosition( 2, '12; 454- 23 $', ['%', '*']) = 0 );

    Check( WordPosition( 0, '12; 454- 23 $', ['2', ' ']) = 0 );
    Check( WordPosition( 1, '12; 454- 23 $', ['2', ' ']) = 1 );
    Check( WordPosition( 2, '12; 454- 23 $', ['2', ' ']) = 3 );
    Check( WordPosition( 3, '12; 454- 23 $', ['2', ' ']) = 5 );
    Check( WordPosition( 4, '12; 454- 23 $', ['2', ' ']) = 11 );
    Check( WordPosition( 5, '12; 454- 23 $', ['2', ' ']) = 13 );
    Check( WordPosition( 6, '12; 454- 23 $', ['2', ' ']) = 0 );
end;

{------------------------------------------------------------------------}

procedure TOdbMiscUnitTest.TestLoadBitmapFromExeRes;
var
    Bitmap : Graphics.TBitmap;
begin
    Bitmap := LoadBitmapFromExeRes('TESTBITMAP');
    try
        Check( Bitmap.Width  = 12 );
        Check( Bitmap.Height = 17 );
        Check( Bitmap.Canvas.Pixels[1,1] = clBlack );
    finally
        Bitmap.Free;
    end;

    Check( LoadBitmapFromExeRes('DOESNOTEXIST') = nil );
end;

{------------------------------------------------------------------------}

{$endif}

initialization

    OdbExceptionProc := DefaultRaiseException;

    { default debug procedures }
    DebugOutputStr := OutputStr;
    DebugOutputInt := OutputInt;
end.
