{----------------------------------------------------------------------------
|
| Library: Spider Containers for Object Pascal
|
| Module: StrTable.Pas
|
| Description: String hash table container.
|
| Copyright (c) 1996,97 Michel Brazeau, Interval Software
|
| History: Version 1.0  March 1996.
|          Dec 11, 1996. MB Added Detach method.
|          Jul 07, 1998. MB Added CreateEx method
|---------------------------------------------------------------------------}
unit StrTable;

{$I Contain.Inc}

interface

uses
    ObjCmp,   { TCompareFunction }
    ObjPerst, { TObjectReader, TObjectWriter }
    ObjType,  { TArrayIndex }
    ObjBase,  { TContainer }
    ObjBuckt, { TStringCombo }
    ObjHash;  { THashTable }

type

{ TStringTable does not inherit from THashTable to not make unecessary
  methods accesible }
TStringTable = class(TContainer)
protected
    function GetSize : LongInt; override;
    function QueryEmpty : Boolean; override;

public
    { By default, the string table is constructed with a case-sensitve
      compare function. }
    constructor Create(const InTableSize : TArrayIndex);

    { Constructor with extended paramaters. InCompareFunc and InHashFunc
      must work with TStringCombo objects. }
    constructor CreateEx( const InTableSize    : TArrayIndex;
                          const InCompareFunc  : TCompareFunc;
                          const InHashFunc     : THashFunc );

    destructor Destroy; override;

    procedure Clear; override;

    { Each string in the table may have a corresponding object. Pass nil
      for Obj if this data is unnecessary }
    procedure Insert(const Str : String; const Obj : TObject);

    { Deletes (removes and Frees) the N'th matching Occur'ence of Str.
      Returns True if the string entry is found and deleted. }
    function Delete( const Str   : String;
                     const Occur : LongInt
                    ) : Boolean;

    { Detaches and returns the N'th matching Occur'ence of Str.
      Returns nil if string not in table. }
    function Detach( const Str   : String;
                     const Occur : LongInt
                    ) : TStringCombo;

    { searched for the N'th occurence of Str. Returns a TStringCombo
      reference if found, else returns nil. }
    function Search( const Str          : String;    { key to search }
                     const Occur        : LongInt    { occurence }
                    ) : TStringCombo;

    { The supplied XXXIterator routine will be passed a TStringCombo
      object for each string in the StringTable }
    procedure ForEachCallProc( const ProcIterator : TProcIterator;
                               const CallBackData : LongInt); override;
    procedure ForEachCallMethod( const MethodIterator : TMethodIterator;
                                 const CallBackData   : LongInt); override;

private
    HashTable : THashTable;

end;

{ Streaming routines }

procedure ReadStringTable( const Obj    : TObject;
                           const Reader : TObjectReader ); far;

procedure WriteStringTable( const Obj    : TObject;
                            const Writer : TObjectWriter ); far;

{--------------------------------------------------------------------------}

implementation

{ hash function for TStringCombo objects. MB Jul 07, 1998. Kept for
 backward compatibity. ObjHash.Pas now has a HashStringCombo function. }
function StringHashFunc(const Obj : TObject; const TableSize : TArrayIndex) : TArrayIndex; far;
begin
    Result := StringHash((Obj as TStringCombo).Str, TableSize);
end;

{ TStringTable implementation ---------------------------------------------}

constructor TStringTable.Create(const InTableSize   : TArrayIndex);
begin
    HashTable := THashTable.Create( TStringCombo, CompareStringCombo,
                                    StringHashFunc, InTableSize);

    inherited Create(TStringCombo);
end;

constructor TStringTable.CreateEx( const InTableSize    : TArrayIndex;
                                   const InCompareFunc  : TCompareFunc;
                                   const InHashFunc     : THashFunc );
begin
    HashTable := THashTable.Create( TStringCombo, InCompareFunc,
                                    InHashFunc, InTableSize);

    inherited Create(TStringCombo);
end;

{--------------------------------------------------------------------------}

destructor TStringTable.Destroy;
begin
    { must call the base destructor first so that HashTable is still valid
      when the call to Clear is made in the base class destructor }
    inherited Destroy;

    HashTable.Free;
end;

{--------------------------------------------------------------------------}

function TStringTable.GetSize : LongInt;
begin
    Result := HashTable.Size;
end;

{--------------------------------------------------------------------------}

function TStringTable.QueryEmpty : Boolean;
begin
    Result := HashTable.IsEmpty;
end;

{--------------------------------------------------------------------------}

procedure TStringTable.Clear;
begin
    { check for nil, as Clear may be invoked on a partially constructed
      object }
    if HashTable <> nil then
        HashTable.Clear;
end;

{--------------------------------------------------------------------------}

procedure TStringTable.Insert(const Str : String; const Obj : TObject);
var
    StringEntry : TStringCombo;
begin
    StringEntry := TStringCombo.Create(Str, Obj);
    try
        HashTable.Insert(StringEntry);
    except
        StringEntry.Free;
        raise;
    end;
end;

{--------------------------------------------------------------------------}

function TStringTable.Delete( const Str   : String;
                              const Occur : LongInt
                             ) : Boolean;
var
    StringKey : TStringCombo;
begin
    StringKey := TStringCombo.Create(Str, nil);
    try
        Delete := HashTable.Delete(StringKey, Occur);
    finally
        StringKey.Free;
    end;
end;

{--------------------------------------------------------------------------}

function TStringTable.Detach( const Str   : String;
                              const Occur : LongInt
                            ) : TStringCombo;
var
    StringKey : TStringCombo;
begin
    StringKey := TStringCombo.Create(Str, nil);
    try
        Detach := HashTable.Detach(StringKey, Occur) as TStringCombo;
    finally
        StringKey.Free;
    end;
end;

{--------------------------------------------------------------------------}

function TStringTable.Search( const Str          : String;
                              const Occur        : LongInt
                             ) : TStringCombo;
var
    Key        : TStringCombo;

begin
    Key := TStringCombo.Create(Str, nil);
    try
        Result := HashTable.Search(Key, Occur) as TStringCombo;
    finally
        Key.Free;
    end;
end;

{--------------------------------------------------------------------------}

procedure TStringTable.ForEachCallProc( const ProcIterator : TProcIterator;
                                        const CallBackData : LongInt);
begin
    HashTable.ForEachCallProc(ProcIterator, CallBackData);
end;

{--------------------------------------------------------------------------}

procedure TStringTable.ForEachCallMethod( const MethodIterator : TMethodIterator;
                                          const CallBackData   : LongInt);
begin
    HashTable.ForEachCallMethod(MethodIterator, CallBackData);
end;

{-------------------------------------------------------------------------}

{ Streaming routines }

procedure ReadStringTable( const Obj    : TObject;
                           const Reader : TObjectReader );
var
    StringTable : TStringTable;

begin
    StringTable := Obj as TStringTable;

    StringTable.HashTable := Reader.ReadChildObject as THashTable;
end;

{-------------------------------------------------------------------------}

procedure WriteStringTable( const Obj    : TObject;
                            const Writer : TObjectWriter );
var
    StringTable : TStringTable;

begin
    StringTable := Obj as TStringTable;

    Writer.WriteChildObject( StringTable.HashTable );
end;

{--------------------------------------------------------------------------}

initialization

    {$ifdef Win32}
    RegisterStreamable( TStringTable,
                        [TObjectReadProc(@ReadStringTable)],
                        [TObjectWriteProc(@WriteStringTable)] );
    {$else}
    RegisterStreamable( TStringTable,
                        [ReadStringTable],
                        [WriteStringTable] );
    {$endif}

    RegisterHashFunction( 'StringHashFunc', StringHashFunc );

end.
